#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

#define NR_POINT_LIGHTS 5
#define PI 3.14159265359

uniform float far_plane;
uniform samplerCube depthMap;
uniform bool shadows;

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

layout (location = 0) out vec4 color;
layout (location = 1) out vec4 brightness;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;

uniform sampler2D albedoMap;
uniform sampler2D normalMap;
uniform sampler2D metallicMap;
uniform sampler2D roughnessMap;
uniform sampler2D aoMap;


uniform sampler2D texture_diffuse1;
uniform sampler2D texture_diffuse2;
uniform sampler2D texture_diffuse3;
uniform sampler2D texture_specular1;
uniform sampler2D texture_specular2;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
};

uniform PointLight pointL[NR_POINT_LIGHTS];

uniform vec3 highlightColor;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

float ShadowCalculation(vec3 fragPos)
{
    // get vector between fragment position and light position
    vec3 fragToLight = fragPos - pointL[0].position;

   // define the closest depth - will be changed in the loop.    
    float closestDepth = texture(depthMap, fragToLight).r;
    
    // now get current linear depth as the length between the fragment and light position
    float currentDepth = length(fragToLight);

    float shadow = 0.0;
    float bias = 0.6;
    int samples = 20;
    float viewDistance = length(camera_world - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        closestDepth = texture(depthMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
        
    // display closestDepth as debug (to visualize depth cubemap)
    color = vec4(vec3(closestDepth / far_plane), 1.0);    
        
    return shadow;
}

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

vec3 getNormalFromMap()
{
    vec3 tangentNormal = texture(normalMap, vert.uv).xyz * 2.0 - 1.0;

    vec3 Q1  = dFdx(vert.position_world);
    vec3 Q2  = dFdy(vert.position_world);
    vec2 st1 = dFdx(vert.uv);
    vec2 st2 = dFdy(vert.uv);

    vec3 N   = normalize(vert.normal_world);
    vec3 T  = normalize(Q1*st2.t - Q2*st1.t);
    vec3 B  = -normalize(cross(N, T));
    mat3 TBN = mat3(T, B, N);

    return normalize(TBN * tangentNormal);
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
	float a = roughness*roughness;
    float a2 = a*a;
    float NdotH = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return a2 / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float denom = NdotV * (1.0 - k) + k;

    return NdotV / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}



vec3 fresnelSchlick(vec3 h, vec3 n, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(max(1.0 - max(dot(h, n), 0.0), 0.0), 5.0);
}  

void main() {	

    vec3 albedo     = pow(texture(albedoMap, vert.uv).rgb, vec3(2.2));
    float metallic  = texture(metallicMap, vert.uv).r;
    float roughness = texture(roughnessMap, vert.uv).r;
    float ao        = texture(aoMap, vert.uv).r;

	//vec3 n = normalize(vert.normal_world);
	vec3 n = getNormalFromMap();
	vec3 v = normalize(camera_world - vert.position_world);
	
	vec3 F0 = vec3(0.04); 
    F0 = mix(F0, albedo, metallic);

	vec3 Lo = vec3(0.0);
	for(int i = 0; i < NR_POINT_LIGHTS; i++){
	  vec3 l = normalize(pointL[i].position - vert.position_world);
	  vec3 h = normalize(v + l);

	  float distance    = length(pointL[i].position - vert.position_world);
	  float attenuation = 1.0 / (pointL[i].attenuation[0] + pointL[i].attenuation[1] * distance + pointL[i].attenuation[2] * (distance * distance));  
	  vec3 radiance = pointL[i].color * attenuation;

	  float NDF = DistributionGGX(n, h, roughness);       
      float G   = GeometrySmith(n, v, l, roughness);  
	  vec3 F    = fresnelSchlick(h, n, F0);

	  vec3 numerator    = NDF * G * F;
      float denominator = 4.0 * max(dot(n, v), 0.0) * max(dot(n, l), 0.0);
      vec3 specular     = numerator / max(denominator, 0.001);  

	  vec3 kS = F;
      vec3 kD = vec3(1.0) - kS;
  
      kD *= 1.0 - metallic;	

	  float nDotl = max(dot(n, l), 0.0);        
	  Lo += (kD * albedo / PI + specular) * radiance * nDotl;
	}

	vec3 ambient = vec3(materialCoefficients.x) * albedo * ao;
	//vec3 ambient = vec3(0.03) * albedo * ao;
    float shadow = shadows ? ShadowCalculation(vert.position_world) : 0.0;  
    vec3 Fcolor   = ambient + (1.0 - shadow) * Lo; 

	Fcolor = Fcolor / (Fcolor + vec3(1.0));
    Fcolor = pow(Fcolor, vec3(1.0/2.2));
	Fcolor += highlightColor;

	color = vec4(Fcolor,1.0);
	
	float bright = dot(Fcolor, vec3(0.7, 0.7, 0.7));
    if(bright > 1.0)
        brightness = color;
    else
        brightness = vec4(0.0, 0.0, 0.0, 1.0);
	
	//vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	//color = vec4(texColor * materialCoefficients.x, 1); // ambient
	
	// add directional light contribution
	//color.rgb += phong(n, -dirL.direction, v, dirL.color * texColor, materialCoefficients.y, dirL.color, materialCoefficients.z, specularAlpha, false, vec3(0));
			
	//add point light contribution
	// color.rgb += phong(n, pointL[0].position - vert.position_world, v, pointL[0].color * texColor, materialCoefficients.y, pointL[0].color, materialCoefficients.z, specularAlpha, true, pointL[0].attenuation);
}

